﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
// type definitions for convenience
using CloneMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItem>;
using SymbolicExpressionTreeNodeItem = HeuristicLab.EvolutionaryTracking.GenericWrapper<HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.SymbolicExpressionTreeNode>;
using TraceMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItemList<HeuristicLab.Core.IItem>>;

namespace HeuristicLab.EvolutionaryTracking {
  /// <summary>
  /// An operator that gathers information on tree fragments that get passed from one generation to the next via genetic operators
  /// (Tracking of genetic variance and heredity)
  /// </summary>
  [Item("SymbolicExpressionTreeFragmentsAnalyzer", "An operator that provides statistics about crossover fragments")]
  [StorableClass]
  public sealed class SymbolicExpressionTreeFragmentsAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string UpdateIntervalParameterName = "UpdateInterval";
    private const string UpdateCounterParameterName = "UpdateCounter";
    private const string ResultsParameterName = "Results";
    private const string SecondaryTraceMapParameterName = "SecondaryTraceMap";
    private const string SecondaryFragmentMapParameterName = "SecondaryFragmentMap";
    private const string SecondaryCloneMapParameterName = "SecondaryCloneMap";
    private const string GlobalTraceMapParameterName = "GlobalTraceMap";
    private const string GlobalCloneMapParameterName = "GlobalCloneMap";
    private const string GlobalFragmentMapParameterName = "GlobalFragmentMap";
    private const string GenerationsParameterName = "Generations";
    private const string FragmentStatisticsParameterName = "FragmentStatistics";
    private const string FragmentLengthsDistributionParametersName = "FragmentLengths";
    private const string FragmentFrequenciesParameterName = "FragmentFrequencies";

    #region Parameter properties
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }
    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    // tracking structures
    public LookupParameter<TraceMapType> GlobalTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[GlobalTraceMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalCloneMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalFragmentMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalFragmentMapParameterName]; }
    }
    public LookupParameter<ItemList<ItemList<IItem>>> FragmentFrequenciesParameter {
      get { return (LookupParameter<ItemList<ItemList<IItem>>>)Parameters[FragmentFrequenciesParameterName]; }
    }
    public LookupParameter<DataTable> FragmentStatisticsParameter {
      get { return (LookupParameter<DataTable>)Parameters[FragmentStatisticsParameterName]; }
    }
    public LookupParameter<DataTable> FragmentLengthsParameter {
      get { return (LookupParameter<DataTable>)Parameters[FragmentLengthsDistributionParametersName]; }
    }
    // auxiliary structures for tracking fragments and individuals from the previous generation
    // (this is needed because tracking is done in connection to the current generation, i.e., for 
    // counting "successful" offspring and fragments
    public LookupParameter<TraceMapType> SecondaryTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[SecondaryTraceMapParameterName]; }
    }
    public LookupParameter<CloneMapType> SecondaryFragmentMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[SecondaryFragmentMapParameterName]; }
    }
    public LookupParameter<CloneMapType> SecondaryCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[SecondaryCloneMapParameterName]; }
    }
    #endregion

    #region Parameters
    public bool EnabledByDefault {
      get { return true; }
    }
    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
    }
    public IntValue UpdateCounter {
      get { return UpdateCounterParameter.Value; }
    }
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public CloneMapType GlobalCloneMap {
      get { return GlobalCloneMapParameter.ActualValue; }
    }
    public TraceMapType GlobalTraceMap {
      get { return GlobalTraceMapParameter.ActualValue; }
    }
    public TraceMapType SecondaryTraceMap {
      get { return SecondaryTraceMapParameter.ActualValue; }
      set { SecondaryTraceMapParameter.ActualValue = value; }
    }
    public CloneMapType SecondaryFragmentMap {
      get { return SecondaryFragmentMapParameter.ActualValue; }
      set { SecondaryFragmentMapParameter.ActualValue = value; }
    }
    public CloneMapType SecondaryCloneMap {
      get { return SecondaryCloneMapParameter.ActualValue; }
      set { SecondaryCloneMapParameter.ActualValue = value; }
    }
    public CloneMapType GlobalFragmentMap {
      get { return GlobalFragmentMapParameter.ActualValue; }
    }
    public IntValue Generations {
      get { return GenerationsParameter.ActualValue; }
    }
    public DataTable FragmentStatistics {
      get { return FragmentStatisticsParameter.ActualValue; }
    }
    public DataTable FragmentLengths {
      get { return FragmentLengthsParameter.ActualValue; }
    }
    public ItemList<ItemList<IItem>> FragmentFrequencies {
      get { return FragmentFrequenciesParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private SymbolicExpressionTreeFragmentsAnalyzer(bool deserializing)
      : base() {
    }
    private SymbolicExpressionTreeFragmentsAnalyzer(SymbolicExpressionTreeFragmentsAnalyzer original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicExpressionTreeFragmentsAnalyzer(this, cloner);
    }
    public SymbolicExpressionTreeFragmentsAnalyzer()
      : base() {
      // analyzer update counter and update interval
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      // tracking
      Parameters.Add(new LookupParameter<TraceMapType>(GlobalTraceMapParameterName, "A global cache containing the whole genealogy."));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalCloneMapParameterName, "A global map keeping track of trees and their clones (made during selection)."));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalFragmentMapParameterName, "A global map keeping track of tree fragments received via crossover."));
      // secondary tracking 
      Parameters.Add(new LookupParameter<TraceMapType>(SecondaryTraceMapParameterName, "Parameter to store the trace map from the previous generation"));
      Parameters.Add(new LookupParameter<CloneMapType>(SecondaryCloneMapParameterName, "Parameter to store the clone map from the previous generation"));
      Parameters.Add(new LookupParameter<CloneMapType>(SecondaryFragmentMapParameterName, "Parameter to store the fragment map from the previous generation"));
      // other params
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations so far."));
      Parameters.Add(new LookupParameter<DataTable>(FragmentStatisticsParameterName, "The data table to store the fragment statistics."));
      Parameters.Add(new LookupParameter<DataTable>(FragmentLengthsDistributionParametersName, "The data table to store the distribution of fragment lengths."));
      Parameters.Add(new LookupParameter<ItemList<ItemList<IItem>>>(FragmentFrequenciesParameterName, "A data structure for fragment frequencies"));
      UpdateCounterParameter.Hidden = true;
      UpdateIntervalParameter.Hidden = true;
    }
    #region After deserialization code
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // check if all the parameters are present and accounted for 
      if (!Parameters.ContainsKey(UpdateIntervalParameterName)) {
        Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      }
      if (!Parameters.ContainsKey(UpdateCounterParameterName)) {
        Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
        UpdateCounterParameter.Hidden = true;
      }
      if (!Parameters.ContainsKey(FragmentLengthsDistributionParametersName)) {
        Parameters.Add(new LookupParameter<DataTable>(FragmentLengthsDistributionParametersName, "The data table to store the distribution of fragment lengths."));
      }
      if (!Parameters.ContainsKey(FragmentFrequenciesParameterName)) {
        Parameters.Add(new LookupParameter<ItemList<ItemList<IItem>>>(FragmentFrequenciesParameterName, "A data structure for fragment frequencies"));
      }
    }
    #endregion
    #region IStatefulItem members
    public override void InitializeState() {
      base.InitializeState();
      UpdateCounter.Value = 0;
    }

    public override void ClearState() {
      base.ClearState();
      UpdateCounter.Value = 0;
    }
    #endregion

    public override IOperation Apply() {
      if (Generations.Value == 0) return base.Apply();
      UpdateCounter.Value++;
      if (UpdateCounter.Value == UpdateInterval.Value) {
        ResultCollection results = ResultsParameter.ActualValue;
        if (FragmentStatistics == null) {
          FragmentStatisticsParameter.ActualValue = new DataTable("Fragment Statistics", "Statistical measurements of fragments aggregated over the whole population");
          FragmentStatistics.VisualProperties.YAxisTitle = "Fragment length/Similarities";
          results.Add(new Result("Fragment Statistics", FragmentStatistics));
        }
        if (FragmentFrequencies == null) {
          FragmentFrequenciesParameter.ActualValue = new ItemList<ItemList<IItem>>();
        }
        if (SecondaryTraceMap == null) {
          SecondaryTraceMap = new TraceMapType();
        }
        if (SecondaryCloneMap == null) {
          SecondaryCloneMap = new CloneMapType();
        }
        if (SecondaryFragmentMap == null) {
          SecondaryFragmentMap = new CloneMapType();
        }

        UpdateCounter.Value = 0; // reset counter
        var gScope = ExecutionContext.Scope;
        while (gScope.Parent != null) gScope = gScope.Parent;

        // for this kind of tracking/analysis, we need at least 2 successive generations of crossover
        if (Generations.Value > 1) {
          #region Fragment Statistics
          InitializeRows();
          // create a dictionary for fast lookup of parents (to see which children from the previous generation become parents, and count their fragments)
          var parentsLookup = new Dictionary<ISymbolicExpressionTree, bool>();
          foreach (var parent in GlobalTraceMap.Values.SelectMany(x => x.Cast<ISymbolicExpressionTree>())) {
            if (!parentsLookup.ContainsKey(parent)) {
              parentsLookup.Add(parent, true);
            }
          }
          // the same, for fast lookup of clones (might not be needed after all)
          var clonesLookup = new Dictionary<ISymbolicExpressionTree, bool>();
          foreach (var original in GlobalCloneMap.Values.Cast<ISymbolicExpressionTree>()) {
            if (!clonesLookup.ContainsKey(original)) {
              clonesLookup.Add(original, true);
            }
          }
          var goodFragments = new List<ISymbolicExpressionTreeNode>();
          var goodParents = new List<ISymbolicExpressionTree>();
          var goodChildren = new List<ISymbolicExpressionTree>();
          // take all individuals that have received a fragment in the previous generation
          foreach (var m in SecondaryFragmentMap) {
            var individual = m.Key as ISymbolicExpressionTree;
            // check if the individual became a parent in the current generation,
            // by checking if it appears among the parents from the current generation trace map
            if (parentsLookup.ContainsKey(individual)) {
              var fragmentItem = m.Value as SymbolicExpressionTreeNodeItem;
              goodFragments.Add(fragmentItem.Content);
              goodParents.AddRange(SecondaryTraceMap[individual].Cast<ISymbolicExpressionTree>());
              goodChildren.Add(individual);
            }
          }
          var allFragments = SecondaryFragmentMap.Values.Select(x => ((SymbolicExpressionTreeNodeItem)x).Content as ISymbolicExpressionTreeNode).ToList();
          var allParents = SecondaryTraceMap.Values.SelectMany(x => x.Cast<ISymbolicExpressionTree>()).ToList();
          var allChildren = SecondaryFragmentMap.Keys.Select(x => x as ISymbolicExpressionTree).ToList();

          FragmentStatistics.Rows["Fragment lengths (good)"].Values.Add(goodFragments.Average(x => x.GetLength()));
          FragmentStatistics.Rows["Fragment lengths (all)"].Values.Add(allFragments.Average(x => x.GetLength()));
          FragmentStatistics.Rows["Parent lengths (good)"].Values.Add(goodParents.Average(x => x.Length));
          FragmentStatistics.Rows["Parent lengths (all)"].Values.Add(allParents.Average(x => x.Length));
          FragmentStatistics.Rows["Child lengths (good)"].Values.Add(goodChildren.Average(x => x.Length));
          FragmentStatistics.Rows["Child lengths (all)"].Values.Add(allChildren.Average(x => x.Length));
          //FragmentStatistics.Rows["Exact similarity (good fragments)"].Values.Add(CalculateSimilarity(goodFragments, (int)SymbolicExpressionTreeMatching.SimilarityLevel.Exact));
          //FragmentStatistics.Rows["Exact similarity (all fragments)"].Values.Add(CalculateSimilarity(allFragments, (int)SymbolicExpressionTreeMatching.SimilarityLevel.Exact));

          //WriteFragmentFrequencies(Environment.GetFolderPath(Environment.SpecialFolder.UserProfile) + "\\FragmentFrequencies.dat");
          #endregion
        }
        // save the current generation so it can be compared to the following one, next time the analyzer is applied 
        SecondaryTraceMap.Clear();
        foreach (var m in GlobalTraceMap)
          SecondaryTraceMap.Add(m.Key, m.Value);
        SecondaryCloneMap.Clear();
        foreach (var m in GlobalCloneMap)
          SecondaryCloneMap.Add(m.Key, m.Value);
        SecondaryFragmentMap.Clear();
        foreach (var m in GlobalFragmentMap)
          SecondaryFragmentMap.Add(m.Key, m.Value);

        // clear the global maps to save memory
        if (this.Successor == null || !(this.Successor is SymbolicExpressionTreeGenealogyAnalyzer)) {
          GlobalCloneMap.Clear();
          GlobalTraceMap.Clear();
          GlobalFragmentMap.Clear();
        }
      }
      return base.Apply();
    }

    private void InitializeRows() {
      if (!FragmentStatistics.Rows.ContainsKey("Parent lengths (good)")) {
        FragmentStatistics.Rows.Add(new DataRow("Parent lengths (good)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Parent lengths (all)")) {
        FragmentStatistics.Rows.Add(new DataRow("Parent lengths (all)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Child lengths (good)")) {
        FragmentStatistics.Rows.Add(new DataRow("Child lengths (good)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Child lengths (all)")) {
        FragmentStatistics.Rows.Add(new DataRow("Child lengths (all)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Fragment lengths (good)")) {
        FragmentStatistics.Rows.Add(new DataRow("Fragment lengths (good)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Fragment lengths (all)")) {
        FragmentStatistics.Rows.Add(new DataRow("Fragment lengths (all)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      // exact similarity
      if (!FragmentStatistics.Rows.ContainsKey("Exact similarity (good fragments)")) {
        FragmentStatistics.Rows.Add(new DataRow("Exact similarity (good fragments)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Exact similarity (all fragments)")) {
        FragmentStatistics.Rows.Add(new DataRow("Exact similarity (all fragments)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
    }

    private void WriteFragmentFrequencies(string path) {
      using (var file = new System.IO.StreamWriter(path)) {
        foreach (var f in FragmentFrequencies) {
          file.WriteLine(((SymbolicExpressionTreeNodeItem)f[0]).Content.GetLength() + " " + ((IntValue)f[1]).Value);
        }
      }
    }

    #region Similarity computations
    /// <summary>
    /// Provide a measure of how similar the fragments that get passed by crossover from one generation to the next are. 
    /// </summary>
    /// <param name="fragments">The symbolic expression tree fragments</param>
    /// <param name="mode">The similarity mode (0 - exact, 1 - high, 2 - relaxed)</param>
    /// <returns>The average number of similar fragments</returns>
    private static double CalculateSimilarity(IList<ISymbolicExpressionTreeNode> fragments, int mode) {
      var visited = new bool[fragments.Count];
      int groups = 0;
      for (int i = 0; i != fragments.Count - 1; ++i) {
        if (visited[i]) continue;
        for (int j = i + 1; j != fragments.Count; ++j) {
          if (visited[j]) continue;
          if (fragments[i].IsSimilarTo(fragments[j], mode)) visited[j] = true;
        }
        ++groups;
      }
      return (double)fragments.Count / groups;
    }
    #endregion
  } //SymbolicExpressionTreeFragmentsAnalyzer
}
